from anndata import AnnData
import scanpy as sc
import numpy as np
import warnings

def filter_on_counts_genes(adata: AnnData, 
                            min_counts: int = 2000, 
                            max_counts: int = 100000, 
                            min_genes: int = 1000, 
                            max_genes: int = 13000, 
                            inplace: bool = False
                        ) -> AnnData | None:
    """
    Filter samples based on counts and genes detected.
    Args:
        adata (.h5ad): Anndata object.
        min_counts (int): Minimum counts detected.
        max_counts (int): Maximum counts detected.
        min_genes (int): Minimum genes detected.
        max_genes (int): Maximum genes detected.
        inplace (bool): Update the adata object in place or return unmodified object with keeper_cells flagged.
    Returns:
        AnnData | None
    """
    if isinstance(adata, AnnData):
        ## Filter cells based on counts and genes detected
        if 'keeper_cells' not in adata.obs.columns:
            adata.obs["keeper_cells"] = [True] * adata.shape[0]
        
        ## Filter cells based on counts detected
        adata.obs["keeper_cells"] &= sc.pp.filter_cells(adata, min_counts = min_counts, inplace=False)[0]
        adata.obs["keeper_cells"] &= sc.pp.filter_cells(adata, max_counts = max_counts, inplace=False)[0]
        
        ## Filter cells based on genes detected
        adata.obs["keeper_cells"] &= sc.pp.filter_cells(adata, min_genes = min_genes, inplace=False)[0]
        adata.obs["keeper_cells"] &= sc.pp.filter_cells(adata, max_genes = max_genes, inplace=False)[0]
        
        ## 
        if inplace:
            adata._inplace_subset_obs(adata.obs["keeper_cells"])
            return None
        else:
            return adata

def filter_on_precomputed_metrics(adata: AnnData, 
                                    doublet_score: float = 0.3, 
                                    pct_counts_mt: float = 3.0, 
                                    GEX_Reads_mapped_confidently_to_genome: float = 0.0, 
                                    GEX_Reads_mapped_to_genome: float = 0.0, 
                                    GEX_Reads_with_TSO: float = 1.0, 
                                    inplace: bool = False,
                                ) -> AnnData | None:
    """
    Filter samples based on precomputed quality control metrics.
    Args:
        adata (.h5ad): Anndata object.
        doublet_score (float): Maximum doublet score.
        pct_counts_mt (float): Maximum percentage of counts in mitochondrial genes.
        GEX_Reads_mapped_confidently_to_genome (float): Minimum percentage of confidently mapped reads. There is no pre-defined good practice for threshold, user must specify.
        GEX_Reads_mapped_to_genome (float): Minimum percentage of reads mapped to genome. There is no pre-defined good practice for threshold, user must specify.
        GEX_Reads_with_TSO (float): Maximum percentage of reads with TSO, per cell. There is no pre-defined good practice for threshold, user must specify.
        inplace (bool): Update the adata object in place or return unmodified object with keeper_cells flagged.
    Returns:
        AnnData | None
    """
    if isinstance(adata, AnnData):
        ## Filter cells based on counts and genes detected
        if 'keeper_cells' not in adata.obs.columns:
            adata.obs["keeper_cells"] = [True] * adata.shape[0]

        ## Gather function parameters
        filter_params = locals(); del filter_params['adata']; del filter_params['inplace']
        
        ## Check that columns are present in the adata object and skip if not present.
        for qc_metric in list(filter_params.keys()):
            if qc_metric not in adata.obs.columns:
                warnings.warn(f"A {qc_metric} column is missing in the adata object. Skipping this metric.", UserWarning)
                del filter_params[qc_metric]

        ## Filter cells based on qc metrics
        if "doublet_score" in filter_params.keys():
            adata.obs["keeper_cells"] &= adata.obs["doublet_score"] < doublet_score
        if "pct_counts_mt" in filter_params.keys():
            adata.obs["keeper_cells"] &= adata.obs["pct_counts_mt"] < pct_counts_mt
        if "GEX_Reads_mapped_confidently_to_genome" in filter_params.keys():
            adata.obs["keeper_cells"] &= adata.obs["GEX_Reads_mapped_confidently_to_genome"] > GEX_Reads_mapped_confidently_to_genome
        if "GEX_Reads_mapped_to_genome" in filter_params.keys():
            adata.obs["keeper_cells"] &= adata.obs["GEX_Reads_mapped_to_genome"] > GEX_Reads_mapped_to_genome
        if "GEX_Reads_with_TSO" in filter_params.keys():
            adata.obs["keeper_cells"] &= adata.obs["GEX_Reads_with_TSO"] < GEX_Reads_with_TSO

        ##
        if inplace:
            adata._inplace_subset_obs(adata.obs["keeper_cells"])
            return None
        else:
            return adata

def filter_utilizing_coarse_labels(adata: AnnData, 
                                    coarse_label_column: str,
                                    coarse_label_map: dict = {"Neurons": ["Excitatory", "Inhibitory"], 
                                                                "Non-Neurons": ["Astrocytes", "Oligodendrocytes", "Microglia", "Endothelial", "Pericytes"]},
                                    coarse_label_threshold: dict = {"Neurons": 2000, "Non-Neurons": 1000},
                                    inplace: bool = False
                                ) -> AnnData | None:
    """
    Filter samples based on coarse labels.
    Args:
        adata (.h5ad): Anndata object.
        coarse_label_column (str): Column name in adata.obs containing coarse labeling identifying neuron and non-neuronal cell types.
        coarse_label_map (dict): Coarse cell type labels to define specific filtering on.
        coarse_label_threshold (dict): Minimum genes detected for each coarse label.
        inplace (bool): Update the adata object in place or return unmodified object with keeper_cells flagged.
    Returns:
        AnnData | None
    """
    if isinstance(adata, AnnData):
        ## Filter cells based on counts and genes detected
        if 'keeper_cells' not in adata.obs.columns:
            adata.obs["keeper_cells"] = [True] * adata.shape[0]
        ## Filter specific quality thresholds for each coarse label
        for coarse_label, coarse_labels in coarse_label_map.items():
            adata.obs["keeper_cells"] &= ~adata.obs[coarse_label_column].isin(coarse_labels) | sc.pp.filter_cells(adata, min_genes = coarse_label_threshold[coarse_label], inplace=False)[0]
        ##
        if inplace:
            adata._inplace_subset_obs(adata.obs["keeper_cells"])
            return None
        else:
            return adata

def filter_on_cluster_entropy(adata: AnnData,
                                cluster_column: str,
                                annotation_columns: list,
                                annotation_thresholds: dict,
                            ) -> AnnData | None:
    """
    Filter samples based on cluster entropy.
    Args:
        adata (.h5ad): Anndata object.
        cluster_column (str): Column name in adata.obs containing cluster labels.
        entropy_columns (str): Column name in adata.obs containing entropy values.
        entropy_thresholds (dict): Minimum entropy values for each annotation being considered.
    Returns:
        AnnData | None
    """
    if isinstance(adata, AnnData):
        ## Filter cells based on counts and genes detected
        if 'keeper_cells' not in adata.obs.columns:
            adata.obs["keeper_cells"] = [True] * adata.shape[0]
        ## Compute cluster entropy for each annotation
        for anno in annotation_columns:
            adata.obs[anno + "_entropy"] = cluster_entropy_qc_metric(adata, cluster_column, anno)
        ## Apply filtering based on entropy thresholds
        for anno, threshold in annotation_thresholds.items():
            adata.obs["keeper_cells"] &= adata.obs[anno + "_entropy"] > threshold
        ##
        if inplace:
            adata._inplace_subset_obs(adata.obs["keeper_cells"])
            return None
        else:
            return adata



from adata import AnnData
import scanpy as sc
import numpy as np

def build_annotation_table(adata: AnnData, 
                            categorical_annotations: list = ["donor_name", "load_name", "roi"],
                            numeric_annotations: list = ["doublet_score", "pct_counts_mt"],
                            min_percent: float = 0.05,
                            annotation_alerts: dict = {"donor_name": 0.90, "load_name": 0.95, "roi": 0.95},
                            add_dominant_library_info: bool = True,
                            add_entropy_info: bool = True,
                            mapping_summary: dict = {}):
    """
    Build a standardized table of annotations describing each cluster to be used for taxonomy development or communication.
    Args:
        adata (.h5ad): Anndata object with `annotations` in adata.obs.
        annotations (list of str): Cluster summary annotations specific to data.
        mapping_summary (dict): A dictionary to store the mapping summary. User can pass in an existing dictionary to append to if desired.
    Returns:
        dict: A mapping_summary containing character and numeric summaries along with donor/lib/roi composition alerts.
    """
    ## Annotations summaries
    for anno in annotations:
        freq = adata.obs.groupby(["cluster", anno])[anno].size()
        mapping_summary[anno] = {}
        for cluster in np.unique(adata.obs.cluster):
            cluster_anno = freq.loc[freq.index.get_level_values("cluster") == cluster].sort_values(ascending=False)
            cluster_anno = np.round(cluster_anno / cluster_anno.sum(), 2)
            record = ""
            for level in cluster_anno.index.get_level_values(anno):
                proportion = cluster_anno.loc[cluster_anno.index.get_level_values(anno) == level].values[0]
                if proportion > min_percent: ## Only show the annotation term when its percentage > 0.05
                    record += level + "(" + str(proportion) + ")" if level == cluster_anno.index.get_level_values(anno)[0] else "| " + level + "(" + str(proportion) + ")"
            mapping_summary[anno][cluster] = record
            if anno in annotation_alerts.keys():
                alert_status = "Balanced" if cluster_anno.max() < annotation_alerts[anno] else "Cautious"
                mapping_summary[anno + "_composition_alert"][cluster] = alert_status

    ## Calculate the median values for basic qc stats
    mapping_summary["cluster_size"] = adata.obs.groupby("cluster").size()
    mapping_summary["gene.counts"] = adata.obs.groupby("cluster")["n_genes_by_counts"].median().round(4)
    mapping_summary["umi.counts"] = adata.obs.groupby("cluster")["n_counts"].median().round(4)
    mapping_summary["umi_gene_ratio"] = ratio.round(4)

    ## Calculate the median values for numeric annotations each cluster
    for col in numeric_annotations:
        mapping_summary[f'{col}_median'] = adata.obs.groupby('cluster')[col].median().round(4)
    
    return mapping_summary

def add_dominant_library_info(adata: AnnData,
                                mapping_summary: dict
                            ) -> dict:
    """
    Add dominant library information to the mapping summary.
    Args:
        adata (.h5ad): Anndata object with `annotations` in adata.obs.
        mapping_summary (dict): A pre-existing mapping summary dictionary to add information on dominant library.
    Returns:
        dict: A mapping_summary containing dominant library information.
    """

    ## Calculate the total number of cells for each library
    library_freq_total = adata.obs.groupby("load_name")["load_name"].size()
    mapping_summary["dominant_library_percentage"] = {}
    mapping_summary["dominant_library_info"] = {}

    ## Add dominant library information to the mapping summary for each cluster
    for cluster in np.unique(adata.obs.cluster):

        ## Calculate the library composition for each cluster
        cluster_library_freq = adata.obs[adata.obs["cluster"] == cluster].groupby("load_name").size()
        cluster_library_normalized = np.round(cluster_library_freq / cluster_library_freq.sum(), 4)      

        ## Information for the dominant library in the cluster
        dominant_library = cluster_library_freq.idxmax()
        dominant_percentage = cluster_library_normalized[dominant_library]
        total_cells_in_dominant_library = library_freq_total[dominant_library]

        ## Update the mapping summary
        mapping_summary["dominant_library_percentage"][cluster] = dominant_percentage
        mapping_summary["dominant_library_info"][cluster] = f"{dominant_library} ({dominant_percentage})% | {total_cells_in_dominant_library}# cells."

    return mapping_summary


from adata import AnnData
import scanpy as sc
import numpy as np

def filter_on_cluster_entropy(adata: AnnData,
                                cluster_column: str,
                                annotation_columns: list,
                                annotation_thresholds: dict,
                            ) -> AnnData | None:
    """
    Filter samples based on cluster entropy.
    Args:
        adata (.h5ad): Anndata object.
        cluster_column (str): Column name in adata.obs containing cluster labels.
        entropy_columns (str): Column name in adata.obs containing entropy values.
        entropy_thresholds (dict): Minimum entropy values for each annotation being considered.
    Returns:
        AnnData | None
    """
    if isinstance(adata, AnnData):
        ## Filter cells based on counts and genes detected
        if 'keeper_cells' not in adata.obs.columns:
            adata.obs["keeper_cells"] = [True] * adata.shape[0]
        ## Compute cluster entropy for each annotation
        for anno in annotation_columns:
            adata.obs[anno + "_entropy"] = cluster_entropy_qc_metric(adata, cluster_column, anno)
        ## Apply filtering based on entropy thresholds
        for anno, threshold in annotation_thresholds.items():
            adata.obs["keeper_cells"] &= adata.obs[anno + "_entropy"] > threshold
        ##
        if inplace:
            adata._inplace_subset_obs(adata.obs["keeper_cells"])
            return None
        else:
            return adata


def cluster_entropy_qc_metric(adata: AnnData,
                                cluster_column: str,
                                annotation_column: str
                            ) -> list:
    """
    Compute entropy (mixing) of an annotation within a pre-defined cluster.
    Args:
        adata (.h5ad): Anndata object.
        cluster_column (str): Column name in adata.obs containing cluster labels.
        annotation_columns (list): Column name in adata.obs containing annotations.
    Returns:
        list: A list of cluster entropy quality control metrics.
    """
    ##
    print("Computing cluster entropy on: " + annotation_column)
    ## Initialize with a value outside range of entropy.
    qc_entropy = [-1] * adata.shape[0]
    ## Compute entropy on annotation column for each cluster
    for cluster in np.unique(adata.obs["cluster_column"]):
        qc_entropy[adata.obs["cluster_column"] == cluster] = \
            scipy.stats.entropy(adata.obs.loc[adata.obs["cluster_column"] == cluster, annotation_column].value_counts() / sum(adata.obs["cluster_column"] == cluster))
    ##
    return qc_entropy

## This is stored on github but needs a better home/repo. 
def cell_entropy_qc_metric(adata: AnnData,
                            annotation_columns: list,
                            nearest_neighbors: int = 15, 
                            dim: str = "X_scVI"
                        ) -> AnnData:
    """Compute entropy (mixing) of annotations within a cells local neighborhood. 
    Args:
        adata (.h5ad): adata object with `annotations` in adata.obs.
        annotations (list of str): Cell level annotations.
        nearest_neighbors (int): Number of nearest neighbors.
        dim (str): Dimensionality reduction within adata.obsm.
    Returns:
        adata
    """
    ## Build nearest neighbor tree for fast lookup
    print("Building nearest neighbor tree.")
    nnTree = KDTree(adata.obsm[dim])
    nearest_dist, nearest_ind = nnTree.query(adata.obsm[dim], k=nearest_neighbors)
    ##
    for anno in annotations:
        print("Computing entropy on: " + anno)
        adata.obs[anno + "_entropy"] = -1 ## Initialize with a value outside range of entropy.
        for cell in range(0, adata.shape[0]):
            nearest_neighbors = nearest_ind[cell,:]
            adata.obs.loc[adata.obs.index[cell], anno + "_entropy"] = scipy.stats.entropy(adata.obs.loc[adata.obs.index[nearest_neighbors],anno].value_counts()/len(nearest_neighbors))
    return adata
